package com.mycompany.myfirstemptybot;

import cz.cuni.amis.introspection.java.JProp;
import cz.cuni.amis.pathfinding.alg.astar.AStarResult;
import cz.cuni.amis.pathfinding.map.IPFMapView;
import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.base3d.worldview.object.ILocated;
import cz.cuni.amis.pogamut.base3d.worldview.object.Location;
import cz.cuni.amis.pogamut.ut2004.agent.module.sensor.AgentInfo;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004Bot;
import cz.cuni.amis.pogamut.ut2004.bot.impl.UT2004BotModuleController;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.SendMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.BotKilled;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.NavPoint;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.Self;
import cz.cuni.amis.pogamut.ut2004.utils.UT2004BotRunner;
import cz.cuni.amis.utils.exception.PogamutException;
import cz.cuni.amis.utils.maps.HashMapSet;
import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * Pogamut "Hello world!" example showing few extra things such as introspection and various bot-initializing methods.
 * <p>
 * We advise you to use simple: DM-TrainingDay map with this example.
 * 
 * <p><p>
 * First, try to run the bot and kill it. If you are a NetBeans user, the bot can be either run by right-clicking the project and choosing Run
 * or by directly running this file. If you are an Eclipse user, run directly this file.
 * 
 * <p><p>
 * Then uncomment the line inside {@link EmptyBot#botKilled(BotKilled)} and run the bot again kill it to see the difference.
 * 
 * <p><p>
 * Then uncomment all lines within {@link EmptyBot#logic()}, run the bot and approach it ... then try to escape from it.
 * 
 * <p><p>
 * Hard to escape? Try the same on map DM-1on1-Albatross
 * 
 * <p><p>
 * If you're using Pogamut NetBeans plugin, try to run the bot, click on its node within Services/UT2004 tab and open properties window (Ctrl+Shift+7).
 * You will see that all fields flagged with {@link JProp} are exported to this window. You can also change their values at runtime.
 * 
 * <p><p>
 * Check out the MAIN method at the end of this file, it contains a hint how to run multiple agents of the same kind at once.
 *
 * @author Jakub Gemrot aka Jimmy
 * @author Michal Bida aka Knight
 * @author Rudolf Kadlec aka ik
 */
@AgentScoped
public class EmptyBot extends UT2004BotModuleController {

    private class PathMapView implements IPFMapView<NavPoint> {

        @Override
        public Collection<NavPoint> getExtraNeighbors(NavPoint node, Collection<NavPoint> mapNeighbors) {
            return null;
        }

        @Override
        public int getNodeExtraCost(NavPoint node, int mapCost) {
            double dist = getMinDistanceFromPaths(node);
            if (dist > 400) return 0;
            return (int)((400 - dist));
        }

        @Override
        public int getArcExtraCost(NavPoint nodeFrom, NavPoint nodeTo, int mapCost) {
            return 0;
        }

        @Override
        public boolean isNodeOpened(NavPoint node) {
            return true;
        }

        @Override
        public boolean isArcOpened(NavPoint nodeFrom, NavPoint nodeTo) {
            return true;
        }
        
    }
    
    private void computeOtherPaths() {
        while (true) {
            AStarResult<NavPoint> result = aStar.findPath(ctf.getOurBase(), ctf.getEnemyBase(), new PathMapView());
            if (!result.isSuccess()) return;
            double avg = getMinAveragePathDistances(result.getPath());
            if (avg < 700) return;
            pathsToEnemyBase.add(result.getPath());
        }
    }

    private double getMinDistanceFromPaths(ILocated location) {
        double result = Double.POSITIVE_INFINITY;
        for (List<NavPoint> path : pathsToEnemyBase) {
            double dist = getMinDistanceFromPath(location.getLocation(), path);
            if (dist < result) result = dist;
        }
        return result;
    }
    
    private double getMinDistanceFromPath(ILocated location, List<NavPoint> path) {
        double result = Double.POSITIVE_INFINITY;
        
        for (NavPoint nav : path) {
            double dist = nav.getLocation().getDistance(location.getLocation());
            if (dist < result) result = dist;
        }
                
        return result;
    }
    
    public double getMinAveragePathDistances(List<NavPoint> path1) {
        double result = Double.POSITIVE_INFINITY;
        for (List<NavPoint> path : pathsToEnemyBase) {
            double dist = getAveragePathDistances(path1, path);
            if (dist < result) result = dist;
        }
        return result;
    }
    
    public double getAveragePathDistances(List<NavPoint> path1, List<NavPoint> path2) {
        double result = 0;
        for (NavPoint nav1 : path1) {
            result += getMinDistanceFromPath(nav1.getLocation(), path2);            
        }
        result /= path1.size();
        return result;
    }

    public enum Place {
    
        HOME,
        MID,
        ENEMY
    
    }   
    
    //
    // Following properties are exported into "Properties" tab inside NetBeans if you have Pogamut NetBeans plugin installed.
    //
    
    @JProp
    public String stringProp = "Hello bot example";
    @JProp
    public boolean boolProp = true;
    @JProp
    public int intProp = 2;
    @JProp
    public double doubleProp = 1.0;  

    /**
     * Initialize all necessary variables here, before the bot actually receives 
     * anything from the environment.
     */
    @Override
    public void prepareBot(UT2004Bot bot) {
        // By uncommenting following line, you can make the bot to do the file logging of all its components
        //bot.getLogger().addDefaultFileHandler(new File("EmptyBot.log"));
    }
    
    static int botCount = 0;
    
    /**
     * Here we can modify initializing command for our bot, e.g., sets its name or skin.
     *
     * @return instance of {@link Initialize}
     */
    @Override
    public Initialize getInitializeCommand() {
        int team = botCount++ % 2 == 0 ? AgentInfo.TEAM_RED : AgentInfo.TEAM_BLUE;
        return new Initialize().setName("EmptyBot").setTeam(team);        
    }

    public HashMapSet<Place, NavPoint> places = new HashMapSet<Place, NavPoint>();
    
    /**
     * Handshake with GameBots2004 is over - bot has information about the map
     * in its world view. Many agent modules are usable since this method is
     * called.
     *
     * @param gameInfo informaton about the game type
     * @param config information about configuration
     * @param init information about configuration
     */
    @Override
    public void botInitialized(GameInfo gameInfo, ConfigChange currentConfig, InitedMessage init) {
      
    }

    /**
     * The bot is initilized in the environment - a physical representation of
     * the bot is present in the game.
     *
     * @param gameInfo informaton about the game type
     * @param config information about configuration
     * @param init information about configuration
     * @param self information about the agent
     */
    @Override
    public void botFirstSpawn(GameInfo gameInfo, ConfigChange config, InitedMessage init, Self self) {
        // Display a welcome message in the game engine
        // right in the time when the bot appears in the environment, i.e., his body has just been spawned 
        // into the UT2004 for the first time.    	
        body.getCommunication().sendGlobalTextMessage("Hello world! I am alive!");

        // alternatively, you may use getAct() method for issuing arbitrary {@link CommandMessage} for the bot's body
        // inside UT2004
        act.act(new SendMessage().setGlobal(true).setText("And I can speak! Hurray!"));

    }
    
    /**
     * This method is called only once right before actual logic() method is
     * called for the first time.
     * 
     * Similar to {@link EmptyBot#botFirstSpawn(cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo, cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange, cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage, cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.Self)}.
     */
    @Override
    public void beforeFirstLogic() {
    }
    
    /**
     * Called each time the bot dies. Good for reseting all bot's state
     * dependent variables.
     *
     * @param event
     */
    @Override
    public void botKilled(BotKilled event) {
        // Uncomment this line to have the bot comment on its death.
        //body.getCommunication().sendGlobalTextMessage("I was KILLED!");
    }
    
    boolean compute = false;
    
    List<List<NavPoint>> pathsToEnemyBase = new ArrayList<List<NavPoint>>();

    /**
     * Main method that controls the bot - makes decisions what to do next. It
     * is called iteratively by Pogamut engine every time a synchronous batch
     * from the environment is received. This is usually 4 times per second - it
     * is affected by visionTime variable, that can be adjusted in ini file
     * inside UT2004/System/GameBots2004.ini
     *
     * @throws cz.cuni.amis.pogamut.base.exceptions.PogamutException
     */
    @Override
    public void logic() throws PogamutException {
        if (compute) {
            compute = false;
            Collection<NavPoint> navPoints = this.navPoints.getNavPoints().values();
            draw.clearAll();
            for (NavPoint nav : navPoints) {
                double distToHome = fwMap.getDistance(nav, ctf.getOurBase());
                double distToEnemy = fwMap.getDistance(nav, ctf.getEnemyBase());
                if (distToHome < distToEnemy) {
                    // HOME - MID
                    if (distToHome / distToEnemy < 0.3) {
                        places.add(Place.HOME, nav);
                        draw.drawCube(info.getTeam() == AgentInfo.TEAM_BLUE ? Color.BLUE : Color.RED, nav.getLocation(), 20);
                    } else {
                        places.add(Place.MID, nav);
                        draw.drawCube(Color.YELLOW, nav.getLocation(), 20);
                    }
                } else {
                    // MID - ENEMY
                    if (distToEnemy / distToHome < 0.3) {
                        places.add(Place.ENEMY, nav);
                        draw.drawCube(info.getTeam() == AgentInfo.TEAM_BLUE ? Color.RED : Color.BLUE, nav.getLocation(), 20);
                    } else {
                        places.add(Place.MID, nav);
                        draw.drawCube(Color.YELLOW, nav.getLocation(), 20);
                    }
                }

            }
            
            pathsToEnemyBase.add(fwMap.getPath(ctf.getOurBase(), ctf.getEnemyBase()));
            
            computeOtherPaths();        

            for (List<NavPoint> path : pathsToEnemyBase) {
                
                NavPoint prev = null;
                for (NavPoint nav : path) {
                    if (prev == null) {
                        prev = nav;
                        continue;
                    }
                    draw.drawLine(Color.GREEN, prev.getLocation(), nav.getLocation());
                    prev = nav;
                }
            }
            
            
        }
    }

    /**
     * This method is called when the bot is started either from IDE or from
     * command line.
     *
     * @param args
     */
    public static void main(String args[]) throws PogamutException {
        new UT2004BotRunner(     // class that wrapps logic for bots executions, suitable to run single bot in single JVM
                EmptyBot.class,  // which UT2004BotController it should instantiate
                "EmptyBot"       // what name the runner should be using
        ).setMain(true)          // tells runner that is is executed inside MAIN method, thus it may block the thread and watch whether agent/s are correctly executed
         .startAgents(4);          // tells the runner to start 1 agent

        // It is easy to start multiple bots of the same class, comment runner above and uncomment following
//        new UT2004BotRunner(EmptyBot.class, "EmptyBot").setMain(true)
//         .startAgents(3);         // tells the runner to start 3 agents at once
    }
}
